@@ -1,44 +1,83 @@
-import unittest
-import unittest.mock
-import test.support
+import contextlib
 import os
 import os.path
-import contextlib
 import sys
+import tempfile
+import test.support
+import unittest
+import unittest.mock
 
 import ensurepip
 import ensurepip._uninstall
 
-# pip currently requires ssl support, so we ensure we handle
-# it being missing (http://bugs.python.org/issue19744)
-ensurepip_no_ssl = test.support.import_fresh_module("ensurepip",
-                                                    blocked=["ssl"])
-try:
-    import ssl
-except ImportError:
-    def requires_usable_pip(f):
-        deco = unittest.skip(ensurepip._MISSING_SSL_MESSAGE)
-        return deco(f)
-else:
-    def requires_usable_pip(f):
-        return f
 
-class TestEnsurePipVersion(unittest.TestCase):
+class TestPackages(unittest.TestCase):
+    def touch(self, directory, filename):
+        fullname = os.path.join(directory, filename)
+        open(fullname, "wb").close()
+
+    def test_version(self):
+        # Test version()
+        with tempfile.TemporaryDirectory() as tmpdir:
+            self.touch(tmpdir, "pip-1.2.3b1-py2.py3-none-any.whl")
+            self.touch(tmpdir, "setuptools-49.1.3-py3-none-any.whl")
+            with (unittest.mock.patch.object(ensurepip, '_PACKAGES', None),
+                  unittest.mock.patch.object(ensurepip, '_WHEEL_PKG_DIR', tmpdir)):
+                self.assertEqual(ensurepip.version(), '1.2.3b1')
+
+    def test_get_packages_no_dir(self):
+        # Test _get_packages() without a wheel package directory
+        with (unittest.mock.patch.object(ensurepip, '_PACKAGES', None),
+              unittest.mock.patch.object(ensurepip, '_WHEEL_PKG_DIR', None)):
+            packages = ensurepip._get_packages()
+
+            # when bundled wheel packages are used, we get _PIP_VERSION
+            self.assertEqual(ensurepip._PIP_VERSION, ensurepip.version())
+
+        # use bundled wheel packages
+        self.assertIsNotNone(packages['pip'].wheel_name)
+        self.assertIsNotNone(packages['setuptools'].wheel_name)
+
+    def test_get_packages_with_dir(self):
+        # Test _get_packages() with a wheel package directory
+        setuptools_filename = "setuptools-49.1.3-py3-none-any.whl"
+        pip_filename = "pip-20.2.2-py2.py3-none-any.whl"
+
+        with tempfile.TemporaryDirectory() as tmpdir:
+            self.touch(tmpdir, setuptools_filename)
+            self.touch(tmpdir, pip_filename)
+            # not used, make sure that it's ignored
+            self.touch(tmpdir, "wheel-0.34.2-py2.py3-none-any.whl")
+
+            with (unittest.mock.patch.object(ensurepip, '_PACKAGES', None),
+                  unittest.mock.patch.object(ensurepip, '_WHEEL_PKG_DIR', tmpdir)):
+                packages = ensurepip._get_packages()
+
+            self.assertEqual(packages['setuptools'].version, '49.1.3')
+            self.assertEqual(packages['setuptools'].wheel_path,
+                             os.path.join(tmpdir, setuptools_filename))
+            self.assertEqual(packages['pip'].version, '20.2.2')
+            self.assertEqual(packages['pip'].wheel_path,
+                             os.path.join(tmpdir, pip_filename))
+
+            # wheel package is ignored
+            self.assertEqual(sorted(packages), ['pip', 'setuptools'])
 
-    def test_returns_version(self):
-        self.assertEqual(ensurepip._PIP_VERSION, ensurepip.version())
 
 class EnsurepipMixin:
 
     def setUp(self):
         run_pip_patch = unittest.mock.patch("ensurepip._run_pip")
         self.run_pip = run_pip_patch.start()
+        self.run_pip.return_value = 0
         self.addCleanup(run_pip_patch.stop)
 
         # Avoid side effects on the actual os module
         real_devnull = os.devnull
         os_patch = unittest.mock.patch("ensurepip.os")
         patched_os = os_patch.start()
+        # But expose os.listdir() used by _find_packages()
+        patched_os.listdir = os.listdir
         self.addCleanup(os_patch.stop)
         patched_os.devnull = real_devnull
         patched_os.path = os.path
@@ -47,13 +86,12 @@ class EnsurepipMixin:
 
 class TestBootstrap(EnsurepipMixin, unittest.TestCase):
 
-    @requires_usable_pip
     def test_basic_bootstrapping(self):
         ensurepip.bootstrap()
 
         self.run_pip.assert_called_once_with(
             [
-                "install", "--no-index", "--find-links",
+                "install", "--no-cache-dir", "--no-index", "--find-links",
                 unittest.mock.ANY, "setuptools", "pip",
             ],
             unittest.mock.ANY,
@@ -62,90 +100,81 @@ class TestBootstrap(EnsurepipMixin, unittest.TestCase):
         additional_paths = self.run_pip.call_args[0][1]
         self.assertEqual(len(additional_paths), 2)
 
-    @requires_usable_pip
     def test_bootstrapping_with_root(self):
         ensurepip.bootstrap(root="/foo/bar/")
 
         self.run_pip.assert_called_once_with(
             [
-                "install", "--no-index", "--find-links",
+                "install", "--no-cache-dir", "--no-index", "--find-links",
                 unittest.mock.ANY, "--root", "/foo/bar/",
                 "setuptools", "pip",
             ],
             unittest.mock.ANY,
         )
 
-    @requires_usable_pip
     def test_bootstrapping_with_user(self):
         ensurepip.bootstrap(user=True)
 
         self.run_pip.assert_called_once_with(
             [
-                "install", "--no-index", "--find-links",
+                "install", "--no-cache-dir", "--no-index", "--find-links",
                 unittest.mock.ANY, "--user", "setuptools", "pip",
             ],
             unittest.mock.ANY,
         )
 
-    @requires_usable_pip
     def test_bootstrapping_with_upgrade(self):
         ensurepip.bootstrap(upgrade=True)
 
         self.run_pip.assert_called_once_with(
             [
-                "install", "--no-index", "--find-links",
+                "install", "--no-cache-dir", "--no-index", "--find-links",
                 unittest.mock.ANY, "--upgrade", "setuptools", "pip",
             ],
             unittest.mock.ANY,
         )
 
-    @requires_usable_pip
     def test_bootstrapping_with_verbosity_1(self):
         ensurepip.bootstrap(verbosity=1)
 
         self.run_pip.assert_called_once_with(
             [
-                "install", "--no-index", "--find-links",
+                "install", "--no-cache-dir", "--no-index", "--find-links",
                 unittest.mock.ANY, "-v", "setuptools", "pip",
             ],
             unittest.mock.ANY,
         )
 
-    @requires_usable_pip
     def test_bootstrapping_with_verbosity_2(self):
         ensurepip.bootstrap(verbosity=2)
 
         self.run_pip.assert_called_once_with(
             [
-                "install", "--no-index", "--find-links",
+                "install", "--no-cache-dir", "--no-index", "--find-links",
                 unittest.mock.ANY, "-vv", "setuptools", "pip",
             ],
             unittest.mock.ANY,
         )
 
-    @requires_usable_pip
     def test_bootstrapping_with_verbosity_3(self):
         ensurepip.bootstrap(verbosity=3)
 
         self.run_pip.assert_called_once_with(
             [
-                "install", "--no-index", "--find-links",
+                "install", "--no-cache-dir", "--no-index", "--find-links",
                 unittest.mock.ANY, "-vvv", "setuptools", "pip",
             ],
             unittest.mock.ANY,
         )
 
-    @requires_usable_pip
     def test_bootstrapping_with_regular_install(self):
         ensurepip.bootstrap()
         self.assertEqual(self.os_environ["ENSUREPIP_OPTIONS"], "install")
 
-    @requires_usable_pip
     def test_bootstrapping_with_alt_install(self):
         ensurepip.bootstrap(altinstall=True)
         self.assertEqual(self.os_environ["ENSUREPIP_OPTIONS"], "altinstall")
 
-    @requires_usable_pip
     def test_bootstrapping_with_default_pip(self):
         ensurepip.bootstrap(default_pip=True)
         self.assertNotIn("ENSUREPIP_OPTIONS", self.os_environ)
@@ -155,7 +184,6 @@ class TestBootstrap(EnsurepipMixin, unittest.TestCase):
             ensurepip.bootstrap(altinstall=True, default_pip=True)
         self.assertFalse(self.run_pip.called)
 
-    @requires_usable_pip
     def test_pip_environment_variables_removed(self):
         # ensurepip deliberately ignores all pip environment variables
         # See http://bugs.python.org/issue19734 for details
@@ -163,7 +191,6 @@ class TestBootstrap(EnsurepipMixin, unittest.TestCase):
         ensurepip.bootstrap()
         self.assertNotIn("PIP_THIS_SHOULD_GO_AWAY", self.os_environ)
 
-    @requires_usable_pip
     def test_pip_config_file_disabled(self):
         # ensurepip deliberately ignores the pip config file
         # See http://bugs.python.org/issue20053 for details
@@ -171,7 +198,7 @@ class TestBootstrap(EnsurepipMixin, unittest.TestCase):
         self.assertEqual(self.os_environ["PIP_CONFIG_FILE"], os.devnull)
 
 @contextlib.contextmanager
-def fake_pip(version=ensurepip._PIP_VERSION):
+def fake_pip(version=ensurepip.version()):
     if version is None:
         pip = None
     else:
@@ -205,7 +232,6 @@ class TestUninstall(EnsurepipMixin, unittest.TestCase):
         self.assertFalse(self.run_pip.called)
 
 
-    @requires_usable_pip
     def test_uninstall(self):
         with fake_pip():
             ensurepip._uninstall_helper()
@@ -217,7 +243,6 @@ class TestUninstall(EnsurepipMixin, unittest.TestCase):
             ]
         )
 
-    @requires_usable_pip
     def test_uninstall_with_verbosity_1(self):
         with fake_pip():
             ensurepip._uninstall_helper(verbosity=1)
@@ -229,7 +254,6 @@ class TestUninstall(EnsurepipMixin, unittest.TestCase):
             ]
         )
 
-    @requires_usable_pip
     def test_uninstall_with_verbosity_2(self):
         with fake_pip():
             ensurepip._uninstall_helper(verbosity=2)
@@ -241,7 +265,6 @@ class TestUninstall(EnsurepipMixin, unittest.TestCase):
             ]
         )
 
-    @requires_usable_pip
     def test_uninstall_with_verbosity_3(self):
         with fake_pip():
             ensurepip._uninstall_helper(verbosity=3)
@@ -253,7 +276,6 @@ class TestUninstall(EnsurepipMixin, unittest.TestCase):
             ]
         )
 
-    @requires_usable_pip
     def test_pip_environment_variables_removed(self):
         # ensurepip deliberately ignores all pip environment variables
         # See http://bugs.python.org/issue19734 for details
@@ -262,7 +284,6 @@ class TestUninstall(EnsurepipMixin, unittest.TestCase):
             ensurepip._uninstall_helper()
         self.assertNotIn("PIP_THIS_SHOULD_GO_AWAY", self.os_environ)
 
-    @requires_usable_pip
     def test_pip_config_file_disabled(self):
         # ensurepip deliberately ignores the pip config file
         # See http://bugs.python.org/issue20053 for details
@@ -271,44 +292,12 @@ class TestUninstall(EnsurepipMixin, unittest.TestCase):
         self.assertEqual(self.os_environ["PIP_CONFIG_FILE"], os.devnull)
 
 
-class TestMissingSSL(EnsurepipMixin, unittest.TestCase):
-
-    def setUp(self):
-        sys.modules["ensurepip"] = ensurepip_no_ssl
-        @self.addCleanup
-        def restore_module():
-            sys.modules["ensurepip"] = ensurepip
-        super().setUp()
-
-    def test_bootstrap_requires_ssl(self):
-        self.os_environ["PIP_THIS_SHOULD_STAY"] = "test fodder"
-        with self.assertRaisesRegex(RuntimeError, "requires SSL/TLS"):
-            ensurepip_no_ssl.bootstrap()
-        self.assertFalse(self.run_pip.called)
-        self.assertIn("PIP_THIS_SHOULD_STAY", self.os_environ)
-
-    def test_uninstall_requires_ssl(self):
-        self.os_environ["PIP_THIS_SHOULD_STAY"] = "test fodder"
-        with self.assertRaisesRegex(RuntimeError, "requires SSL/TLS"):
-            with fake_pip():
-                ensurepip_no_ssl._uninstall_helper()
-        self.assertFalse(self.run_pip.called)
-        self.assertIn("PIP_THIS_SHOULD_STAY", self.os_environ)
-
-    def test_main_exits_early_with_warning(self):
-        with test.support.captured_stderr() as stderr:
-            ensurepip_no_ssl._main(["--version"])
-        warning = stderr.getvalue().strip()
-        self.assertTrue(warning.endswith("requires SSL/TLS"), warning)
-        self.assertFalse(self.run_pip.called)
-
 # Basic testing of the main functions and their argument parsing
 
-EXPECTED_VERSION_OUTPUT = "pip " + ensurepip._PIP_VERSION
+EXPECTED_VERSION_OUTPUT = "pip " + ensurepip.version()
 
 class TestBootstrappingMainFunction(EnsurepipMixin, unittest.TestCase):
 
-    @requires_usable_pip
     def test_bootstrap_version(self):
         with test.support.captured_stdout() as stdout:
             with self.assertRaises(SystemExit):
@@ -317,13 +306,12 @@ class TestBootstrappingMainFunction(EnsurepipMixin, unittest.TestCase):
         self.assertEqual(result, EXPECTED_VERSION_OUTPUT)
         self.assertFalse(self.run_pip.called)
 
-    @requires_usable_pip
     def test_basic_bootstrapping(self):
-        ensurepip._main([])
+        exit_code = ensurepip._main([])
 
         self.run_pip.assert_called_once_with(
             [
-                "install", "--no-index", "--find-links",
+                "install", "--no-cache-dir", "--no-index", "--find-links",
                 unittest.mock.ANY, "setuptools", "pip",
             ],
             unittest.mock.ANY,
@@ -331,6 +319,13 @@ class TestBootstrappingMainFunction(EnsurepipMixin, unittest.TestCase):
 
         additional_paths = self.run_pip.call_args[0][1]
         self.assertEqual(len(additional_paths), 2)
+        self.assertEqual(exit_code, 0)
+
+    def test_bootstrapping_error_code(self):
+        self.run_pip.return_value = 2
+        exit_code = ensurepip._main([])
+        self.assertEqual(exit_code, 2)
+
 
 class TestUninstallationMainFunction(EnsurepipMixin, unittest.TestCase):
 
@@ -342,10 +337,9 @@ class TestUninstallationMainFunction(EnsurepipMixin, unittest.TestCase):
         self.assertEqual(result, EXPECTED_VERSION_OUTPUT)
         self.assertFalse(self.run_pip.called)
 
-    @requires_usable_pip
     def test_basic_uninstall(self):
         with fake_pip():
-            ensurepip._uninstall._main([])
+            exit_code = ensurepip._uninstall._main([])
 
         self.run_pip.assert_called_once_with(
             [
@@ -354,6 +348,13 @@ class TestUninstallationMainFunction(EnsurepipMixin, unittest.TestCase):
             ]
         )
 
+        self.assertEqual(exit_code, 0)
+
+    def test_uninstall_error_code(self):
+        with fake_pip():
+            self.run_pip.return_value = 2
+            exit_code = ensurepip._uninstall._main([])
+        self.assertEqual(exit_code, 2)
 
 
 if __name__ == "__main__":
